import unittest
from datetime import datetime, timedelta

import pytz

from ..models import LocalizedRecurrence, LocalizedRecurrenceQuerySet, replace_with_offset


class Test_LocalizedRecurrenceQuerySet(unittest.TestCase):
    """Simple test to ensure the custom query set is being used.
    """
    def setUp(self):
        self.lr_day = LocalizedRecurrence.objects.create(
            interval='DAY',
            offset=timedelta(hours=12),
            timezone=pytz.timezone('US/Eastern'),
        )

    def test_isinstance(self):
        recurrences = LocalizedRecurrence.objects.all()
        self.assertIsInstance(recurrences, LocalizedRecurrenceQuerySet)


class Test_LocalizedRecurrenceQuerySet_update_schedule(unittest.TestCase):
    """Test that updates to recurrences are reflected in the DB.
    """
    def setUp(self):
        self.lr_day = LocalizedRecurrence.objects.create(
            interval='DAY',
            offset=timedelta(hours=12),
            timezone=pytz.timezone('US/Eastern'),
        )
        self.lr_month = LocalizedRecurrence.objects.create(
            interval='MONTH',
            offset=timedelta(hours=15),
            timezone=pytz.timezone('US/Eastern'),
        )

    def test_update_from_1970(self):
        """Start with next_scheduled of 1970, after update should be new.
        """
        time = datetime(year=2013, month=5, day=20, hour=12, minute=3)
        LocalizedRecurrence.objects.filter(interval='DAY').update_schedule(time=time)
        lr_day = LocalizedRecurrence.objects.filter(interval='DAY').first()
        self.assertGreater(lr_day.next_scheduled, time)

    def test_update_all(self):
        """Calls to the model manager to update should be passed through.
        """
        time = datetime(year=2013, month=5, day=20, hour=15, minute=3)
        LocalizedRecurrence.objects.update_schedule(time=time)
        self.assertTrue(all(r.next_scheduled > time for r in LocalizedRecurrence.objects.all()))


class Test_LocalizedRecurrence(unittest.TestCase):
    """Test the creation and querying of LocalizedRecurrence records.
    """
    def setUp(self):
        self.lr = LocalizedRecurrence.objects.create(
            interval='DAY',
            offset=timedelta(hours=12),
            timezone=pytz.timezone('US/Eastern'),
        )

    def test_creation_works(self):
        """Create a single record and get it back in a queryset.

        The record is created in setUp, we just test the return here.
        """
        lr_count = LocalizedRecurrence.objects.all().count()
        self.assertEqual(lr_count, 1)

    def test_timedelta_returned(self):
        """Test that the Duration field is correctly returning timedeltas.
        """
        lr = LocalizedRecurrence.objects.first()
        self.assertTrue(isinstance(lr.offset, timedelta))


class Test_LocalizedRecurrence_utc_of_next_schedule(unittest.TestCase):
    def setUp(self):
        self.lr_day = LocalizedRecurrence.objects.create(
            interval='DAY',
            offset=timedelta(hours=12),
            timezone=pytz.timezone('US/Eastern'),
        )

        self.lr_week = LocalizedRecurrence.objects.create(
            interval='WEEK',
            offset=timedelta(days=3, hours=17, minutes=30),
            timezone=pytz.timezone('US/Central')
        )

        self.lr_month = LocalizedRecurrence.objects.create(
            interval='MONTH',
            offset=timedelta(days=21, hours=19, minutes=15, seconds=10),
            timezone=pytz.timezone('US/Central')
        )

    def test_basic_works(self):
        """Test a simple case of utc_of_next_schedule.

        - The given recurrence is scheduled daily for Eastern Time at noon.
        - The given current date in UTC is 2013/1/15::17:05:22
        - We then expect the next schedule in UTC to be 2013/1/16::17:0:0
        """
        current_time = datetime(2013, 1, 15, 17, 5, 22)
        expected_next_schedule = datetime(2013, 1, 16, 17)
        schedule_out = self.lr_day.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_before_midnight(self):
        """The case when the scheduled and current time cross midnight.

        - The given recurrence is scheduled daily for Eastern Time at 11:59PM.
        - The given current date in UTC is 2013/1/15::05:05:22 -> 12:05AM EST
        - We then expect the next schedule in UTC to be 2013/1/16::05:0:0
        """
        self.lr_day.offset = timedelta(hours=23, minutes=59)
        self.lr_day.save()
        current_time = datetime(2013, 1, 15, 5, 5, 22)
        expected_next_schedule = datetime(2013, 1, 16, 4, 59)
        schedule_out = self.lr_day.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_after_midnight(self):
        """The case when the scheduled and current time are after midnight.

        - The given recurrence is scheduled daily for Eastern Time at 12:01 AM.
        - The given current date in UTC is 2013/1/15::05:05:22 -> 12:05AM EST
        - We then expect the next schedule in UTC to be 2013/1/16::05:01:0
        """
        self.lr_day.offset = timedelta(minutes=1)
        self.lr_day.save()
        current_time = datetime(2013, 1, 15, 5, 5, 22)
        expected_next_schedule = datetime(2013, 1, 16, 5, 01)
        schedule_out = self.lr_day.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_week_update(self):
        """Weekly Recurrences should work as expected.

        - Friday August 8th at 10:34 PM UTC is 5:34 PM CDT.
        - Scheduled for weekly, Friday at 5:30
        - Expect next schedule to be Friday August 15th at 10:30 PM UTC
        """
        current_time = datetime(2013, 8, 8, 22, 34)
        expected_next_schedule = datetime(2013, 8, 15, 22, 30)
        schedule_out = self.lr_week.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_month_update(self):
        """Monthly Recurrences should work as expected.

        - Friday August 23rd at 12:34 AM UTC is Friday August 22nd at 7:34 PM CDT.
        - Scheduled for Monthly, on the 8th day at 7:15.10 PM CDT
        - Expect next schedule to be September 9 at 12:15.10 AM UTC
        """
        current_time = datetime(2013, 8, 23, 0, 34, 55)
        expected_next_schedule = datetime(2013, 9, 23, 0, 15, 10)
        schedule_out = self.lr_month.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_into_dst_boundary(self):
        """Recurrences happen at the correct local time after going into DST.

        Going into daylight savings time should mean the UTC time is
        an hour less on the next recurrence.

        - 2013 US DST began: Sunday March 10th.
        - Weekly recurence, every Thursday at Noon EST.
        """
        self.lr_week.offset = timedelta(days=3, hours=12)
        self.lr_week.save()
        current_time = datetime(2013, 3, 7, 18)
        expected_next_schedule = datetime(2013, 3, 14, 17)
        schedule_out = self.lr_week.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_out_of_dst_boundary(self):
        """Recurrences at the correct local time after going out of UTC.

        Going into daylight savings time should mean the UTC time is
        an hour greater on the next recurrence.

        - 2013 US DST ended: Sunday November 3rd.
        - Weekly recurence, every Thursday at Noon EST.
        """
        self.lr_week.offset = timedelta(days=3, hours=12)
        self.lr_week.save()
        current_time = datetime(2013, 10, 31, 17)
        expected_next_schedule = datetime(2013, 11, 7, 18)
        schedule_out = self.lr_week.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)

    def test_utc_plus(self):
        """Test a timezone that is UTC + 2.

        - Europe/Berlin DST is UTC + 2
        """
        self.lr_day.timezone = pytz.timezone('Europe/Berlin')
        self.lr_day.save()
        current_time = datetime(2013, 5, 5, 10, 10)
        expected_next_schedule = datetime(2013, 5, 6, 10)
        schedule_out = self.lr_day.utc_of_next_schedule(current_time)
        self.assertEqual(schedule_out, expected_next_schedule)


class Test_replace_with_offset(unittest.TestCase):
    def test_day(self):
        """replace_with_offset works as expected with a 'DAY' interval.
        """
        dt_in = datetime(2013, 1, 20, 12, 45, 48)
        td_in = timedelta(hours=3, minutes=3, seconds=3)
        interval_in = 'DAY'
        dt_expected = datetime(2013, 1, 20, 3, 3, 3)
        dt_out = replace_with_offset(dt_in, td_in, interval_in)
        self.assertEqual(dt_out, dt_expected)

    def test_week(self):
        """replace_with_offset works as expected with a 'WEEK' interval.
        """
        dt_in = datetime(2013, 1, 20, 12, 45, 48)
        td_in = timedelta(days=4, hours=3, minutes=3, seconds=3)
        interval_in = 'WEEK'
        dt_expected = datetime(2013, 1, 18, 3, 3, 3)
        dt_out = replace_with_offset(dt_in, td_in, interval_in)
        self.assertEqual(dt_out, dt_expected)

    def test_week_on_month_boundary(self):
        """replace_with_offset using interval 'WEEK' should roll over months
        correctly.
        """
        dt_in = datetime(2013, 7, 30, 12, 45, 48)
        td_in = timedelta(days=4, hours=3, minutes=3, seconds=3)
        interval_in = 'WEEK'
        dt_expected = datetime(2013, 8, 2, 3, 3, 3)
        dt_out = replace_with_offset(dt_in, td_in, interval_in)
        self.assertEqual(dt_out, dt_expected)

    def test_month(self):
        """replace_with_offset works as expected with a 'MONTH' interval.
        """
        dt_in = datetime(2013, 1, 20, 12, 45, 48)
        td_in = timedelta(days=15, hours=3, minutes=3, seconds=3)
        interval_in = 'MONTH'
        dt_expected = datetime(2013, 1, 16, 3, 3, 3)
        dt_out = replace_with_offset(dt_in, td_in, interval_in)
        self.assertEqual(dt_out, dt_expected)

    def test_bad_interval(self):
        """A missformed interval should raise a value error
        """
        dt_in = datetime(2013, 1, 20, 12, 45, 48)
        td_in = timedelta(days=15, hours=3, minutes=3, seconds=3)
        interval_in = 'blah'
        with self.assertRaises(ValueError):
            replace_with_offset(dt_in, td_in, interval_in)



from datetime import datetime, timedelta

from dateutil.relativedelta import relativedelta
from django.db import models
from timezone_field import TimeZoneField
import fleming
import pytz

from .fields import DurationField


INTERVAL_CHOICES = (
    ('DAY', 'Day'),
    ('WEEK', 'Week'),
    ('MONTH', 'Month')
)


class LocalizedRecurrenceQuerySet(models.query.QuerySet):
    def update_schedule(self, time=None):
        """Update the schedule times for all the recurrences in the queryset.
        """
        time = time or datetime.utcnow()
        for recurrence in self:
            recurrence.next_scheduled = recurrence.utc_of_next_schedule(time)
            recurrence.previous_scheduled = time
            recurrence.save()


class LocalizedRecurrenceManager(models.Manager):
    def get_queryset(self):
        return LocalizedRecurrenceQuerySet(self.model)

    def __getattr__(self, name):
        """
        Written to allow both:
            - LocalizedRecurrence.objects.update_schedule()
            - LocalizedRecurrence.get(id=my_recurrence).update_schedule()
        """
        return getattr(self.get_queryset(), name)


class LocalizedRecurrence(models.Model):
    """The information necessary to act on events in users local times.
    """
    interval = models.CharField(max_length=18, default='DAY', choices=INTERVAL_CHOICES)
    offset = DurationField(default=timedelta(0))
    timezone = TimeZoneField(default='UTC')
    previous_scheduled = models.DateTimeField(default=datetime(1970, 1, 1))
    next_scheduled = models.DateTimeField(default=datetime(1970, 1, 1))

    objects = LocalizedRecurrenceManager()

    def utc_of_next_schedule(self, current_time):
        local_time = fleming.convert_to_tz(current_time, self.timezone)
        local_scheduled_time = replace_with_offset(local_time, self.offset, self.interval)
        utc_scheduled_time = fleming.convert_to_tz(local_scheduled_time, pytz.utc, return_naive=True)
        if utc_scheduled_time <= current_time:
            additional_time = {
                'DAY': timedelta(days=1),
                'WEEK': timedelta(weeks=1),
                'MONTH': relativedelta(months=1)
            }
            utc_scheduled_time = fleming.add_timedelta(
                utc_scheduled_time, additional_time[self.interval], within_tz=self.timezone)
        return utc_scheduled_time


def replace_with_offset(dt, offset, interval):
    """Replace components of a datetime with those of a timedelta.

    This replacement is done within the given interval. This means the
    the final result, will the be a datetime, at the desired offset
    given the interval.
    """
    hours, minutes, seconds = offset.seconds // 3600, (offset.seconds // 60) % 60, offset.seconds % 60
    interval = interval.lower()
    if interval == 'day':
        dt_out = dt.replace(hour=hours, minute=minutes, second=seconds)
    elif interval == 'week':
        dt_out = dt + timedelta(days=offset.days - dt.weekday())
        dt_out = dt_out.replace(hour=hours, minute=minutes, second=seconds)
    elif interval == 'month':
        # TODO:
        #     - Modify so it works with the last day of the month
        #     - As per: http://stackoverflow.com/questions/42950/get-last-day-of-the-month-in-python
        #     - Add test for: e.g. February 30th.
        dt_out = dt.replace(day=offset.days + 1, hour=hours, minute=minutes, second=seconds)
    else:
        raise ValueError('{i} is not a proper interval value'.format(i=interval))
    return dt_out



import unittest
from datetime import timedelta

from mock import MagicMock, patch

from .. import fields


class Test_DurationField_to_python(unittest.TestCase):
    def setUp(self):
        """Create a mock of the DurationField class and pin to_python to it.
        """
        self.df = fields.DurationField()

    def test_timedelta(self):
        """A timedelta should just get returned.
        """
        td_in = timedelta(days=1, hours=3)
        td_out = self.df.to_python(td_in)
        self.assertEqual(td_out, td_in)

    def test_string(self):
        """A string should be properly converted.
        """
        td_expected = timedelta(days=3, hours=5, minutes=10)
        str_in = str(td_expected)
        td_out = self.df.to_python(str_in)
        self.assertEqual(td_out, td_expected)

    def test_bad_string(self):
        """Malformed strings should raise an error.
        """
        str_in = "3:5"
        with self.assertRaises(ValueError):
            self.df.to_python(str_in)

    def test_int(self):
        """Int inputs should give timedelta outputs
        """
        int_in = 176400
        td_expected = timedelta(days=2, hours=1)
        td_out = self.df.to_python(int_in)
        self.assertEqual(td_out, td_expected)

    def test_none(self):
        """Null input -> None output
        """
        td_in = None
        td_out = self.df.to_python(td_in)
        self.assertEqual(td_in, td_out)

    def test_invalid(self):
        """Unsupported objects raise ValueError
        """
        td_in = self
        with self.assertRaises(ValueError):
            self.df.to_python(td_in)


class Test_DurationField_get_prep_value(unittest.TestCase):
    def setUp(self):
        """Create a mock of the DurationField class and pin get_prep_value.
        """
        self.df = fields.DurationField()

    def test_returns_int(self):
        """Type should be `int`.
        """
        td_in = timedelta(hours=14)
        int_out = self.df.get_prep_value(td_in)
        self.assertTrue(isinstance(int_out, int))

    def test_returns_int2(self):
        """Type should be `int`.
        """
        td_in = timedelta(hours=14, seconds=3.5)
        int_out = self.df.get_prep_value(td_in)
        self.assertTrue(isinstance(int_out, int))

    def test_round_trip(self):
        """A trip through get_prep_value and to_python.
        """
        td_in = timedelta(hours=14)
        int_out = self.df.get_prep_value(td_in)
        td_out = self.df.to_python(int_out)
        self.assertEqual(td_out, td_in)


class Test_DurationField_value_to_string(unittest.TestCase):
    def setUp(self):
        """Create a mock of the DurationField class and pin value_to_string to it.

        We also mock out the _get_val_from_obj, to just return what is
        passed in. Allows us to pass in a timedelta for testing intead
        of the expected object.
        """
        self.mock_DurationField = MagicMock()
        self.mock_DurationField.value_to_string = fields.DurationField.__dict__['value_to_string']
        self.mock_DurationField.to_python = fields.DurationField.__dict__['to_python']
        self.mock_DurationField._get_val_from_obj.side_effect = lambda x: x

    def test_simple_string(self):
        """
        """
        in_td = timedelta(days=1, hours=1, minutes=1, seconds=1)
        expected_str = "1 day, 1:01:01"
        out_str = self.mock_DurationField.value_to_string(self.mock_DurationField, in_td)
        self.assertEqual(out_str, expected_str)

    def test_loop_to_python(self):
        """A roundtrip through value_to_string and to_python.

        We expect the value to be unchanged after the trip.
        """
        in_td = timedelta(days=1, hours=1, minutes=1, seconds=1)
        out_str = self.mock_DurationField.value_to_string(self.mock_DurationField, in_td)
        out_td = self.mock_DurationField.to_python(self.mock_DurationField, out_str)
        self.assertEqual(out_td, in_td)


class Test_parse_timedelta_string(unittest.TestCase):
    """Test some round trips between timedelta, string, timedelta.
    """
    def test_works(self):
        """Test a bunch of cases for the regular expressions.

        We test so many things in this test because regular
        expressions represent a lot of effective branching.
        """
        td1_in = timedelta(days=1)
        td2_in = timedelta(seconds=1)
        td3_in = timedelta(hours=1, minutes=1, seconds=1)
        td4_in = timedelta(days=1, seconds=10000.001)
        td5_in = timedelta(hours=12)
        td6_in = timedelta(hours=-12)
        str1_out = str(td1_in)
        str2_out = str(td2_in)
        str3_out = str(td3_in)
        str4_out = str(td4_in)
        str5_out = str(td5_in)
        str6_out = str(td6_in)
        td1_out = fields.parse_timedelta_string(str1_out)
        td2_out = fields.parse_timedelta_string(str2_out)
        td3_out = fields.parse_timedelta_string(str3_out)
        td4_out = fields.parse_timedelta_string(str4_out)
        td5_out = fields.parse_timedelta_string(str5_out)
        td6_out = fields.parse_timedelta_string(str6_out)
        self.assertEqual(td1_in, td1_out)
        self.assertEqual(td2_in, td2_out)
        self.assertEqual(td3_in, td3_out)
        self.assertEqual(td4_in, td4_out)
        self.assertEqual(td5_in, td5_out)
        self.assertEqual(td6_in, td6_out)


class Test_setup_south(unittest.TestCase):
    def test_no_south(self):
        """This test is meant to hit the branch handleing the ImportError for
        south.

        Some users may not have south, so if it is not available, we
        just pass.
        """
        with patch.dict('sys.modules', {'south.modelsinspector': {}}):
            reload(fields)



from datetime import timedelta
import re

from django.db.models import SubfieldBase
from django.db.models.fields import IntegerField


class DurationField(IntegerField):
    """A field to store durations of time with accuracy to the second.

    Automatically converts between python timedelta objects and
    database integers.

    The timedeltas are stored as seconds in the backend, so sub-second
    accuracy is lost with this field.
    """
    description = "A duration of time."
    __metaclass__ = SubfieldBase

    def to_python(self, value):
        """Convert a stored duration into a python datetime.timedelta object.

        Need to handle three cases:
            - An instance of the correct type (timedelta).
            - A string (e.g., from a deserializer).
            - What the database returns (integer).
        """
        if isinstance(value, timedelta):
            v = value
        elif isinstance(value, (str, unicode)):
            # The string should be in the form "[D day[s],][H]H:MM:SS[.UUUUUU]"
            try:
                v = parse_timedelta_string(value)
            except ValueError:
                raise ValueError("Duration string must be in the form '[D day[s],][H]H:MM:SS[.UUUUUU]'")
        elif isinstance(value, int):
            v = timedelta(seconds=value)
        elif value is None:
            v = None
        else:
            raise ValueError("Not a valid Duration object")
        return v

    def get_prep_value(self, value):
        """Convert a timedelta to integer for storage.
        """
        # import pdb; pdb.set_trace()
        value = self.to_python(value)
        return int(value.total_seconds())

    def value_to_string(self, obj):
        """Used by serializers to get a string representation.
        """
        time_delta_value = self._get_val_from_obj(obj)
        return time_delta_value.__str__()


def parse_timedelta_string(string):
    """Parses strings from datetime.timedelta.__str__.

    Usefull because django's force_text uses .__str__ instead of
    .__repr__

    datetime.timedelta.__str__ returns a string in the form [D day[s],
    ][H]H:MM:SS[.UUUUUU], where D is negative for negative t.
    """
    days_re = "(?P<days>-?[0-9]*) days?, (?P<hours>[0-9]+):(?P<minutes>[0-9]+):(?P<seconds>[0-9]+\.?[0-9]*)"
    no_days_re = "(?P<hours>[0-9]+):(?P<minutes>[0-9]+):(?P<seconds>[0-9]+\.?[0-9]*)"
    match_days = re.match(days_re, string)
    match_no_days = re.match(no_days_re, string)
    if match_days:
        return timedelta(**{k: float(v) for k, v in match_days.groupdict().items()})
    elif match_no_days:
        return timedelta(**{k: float(v) for k, v in match_no_days.groupdict().items()})
    else:
        raise ValueError("'%s' is not in the form [D day[s],][H]H:MM:SS[.UUUUUU]" % string)


try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ["^localized_recurrence\.fields\.DurationField"])
except ImportError:
    pass
